from . import raw_nodes, schema, utils
from .raw_nodes import FormatVersion

try:
    from typing import get_args
except ImportError:
    from typing_extensions import get_args  # type: ignore

format_version = get_args(FormatVersion)[-1]
